*-------------------------------------------------------------------------------
*- Synopsis...: This program compares 2 dbf (or SQL) tables record by record and
*-              in each record the fields in the first file are compared
*-              with the second file. Reports any differences in an
*-              output file.
*-
*-              Compares only non deleted records.
*-              Runs only in Visual FoxPro 6.0 & above.
*-              Ignores fields that are there in the primary file but
*-                not in the secondary file or if their type do not match.
*-              Ignores errors with General fields.
*-
*-              By setting the parameter pl_FileAlias as .t., you can do the
*-              compare on files that are already open. In this case the filter
*-              conditions & orders set for the files would be used. In this
*-              case the tables are kept open when exiting this program.
*-              You can use this to compare one VFP table with its SQL 
*-              counter part
*-
*-              SQL Tables: You can use this tool to compare 2 SQL tables. To do
*-                          this, first establish a connection to the server by
*-                          using SQLEXEC() or DO tools\Sqltest. Then pass that
*-                          handle to pn_SqlHandle parameter. For table name
*-                          parameters like pc_File1, pc_File2, you can 
*-                          use <datbase name>..<table name> 
*-                          You could also have the tables open before 
*-                          (either through remote view or SQLEXEC()) and then
*-                          use their alias for comparing.
*-
*-              Will not give any errors if the tables are totally different.
*-------------------------------------------------------------------------------
*- Parameters.: pc_File1, The primary table to compare.
*-              pc_File2, The other table to compare with.
*-              pc_Description, Description of the comparison [Optional]
*-              pl_FileAlias, If .t., the first 2 parameters are the
*-                            aliases and not file names.
*-              pc_OutputFile, The file to output to. If not passed or is
*-                             empty will prompt the user to give the file
*-              pc_RecordExpr, This expression will be evaluated and added to the file
*-                             whenever the values do not match.
*-              pc_IgnoreFieldList, If passed, should be a comma delimited list
*-                                  of fields to ignore in the compare
*-              pl_NoEraseOutputFile, If .t., will not erase output file
*-                                    if it already exists. Instead will add
*-                                    the errors to that file.
*-              pl_NoDisplayErrors, If .f. will MODI FILE the output file
*-                                  if there were errors.
*-                                  If .t. avoids this.
*-              pl_TimeStampCompare, If .f. will not compare adduser, adddate,
*-                                   addtime and lckuser, lckdate, lcktime.
*-                                   If .t. will do the compare for these
*-                                   fields also.
*-              pn_SqlHandle,       If passed, assumes the table are SQL server tables
*-                                  In this case, it uses SQLEXEC() to get all 
*-                                  the records to a temp VFP cursor and then 
*-                                  works from there. 
*-                                  Here you can use pc_File1
*-                                  to be <database1>..<table>
*-                                  and pc_File2 to be <database2>..<table>
*-                                  to compare across 2 databases.
*-------------------------------------------------------------------------------
*- Returns....: <expC>, The name of the output file if the process was
*-                      successful, else blank string
*-------------------------------------------------------------------------------
LPARAMETERS pc_File1, ;
            pc_File2, ;
            pc_Description, ;
            pl_FileAlias, ;
            pc_OutputFile, ;
            pc_RecordExpr, ;
            pc_IgnoreFieldList, ;
            pl_NoEraseOutputFile, ;
            pl_NoDisplayErrors, ;
            pl_TimeStampCompare, ;
            pn_SqlHandle

#define Cc_Null              SPACE(0)
#define Cc_OsEol             CHR(13) + CHR(10)

LOCAL lc_File1, ;
      lc_File2, ;
      lc_OutputFile, ;
      ll_Cont, ;
      lc_Deleted, ;
      ln_Count, ;
      ln_Remove, ;
      lc_Field1, ;
      lc_Field2, ;
      lc_Alias1, ;
      lc_Alias2, ;
      ll_FirstRec, ;
      ln_ErrorCount, ;
      lc_IgnoreFieldList, ;
      lc_Description

LOCAL ARRAY la_Fields[1]

PUSH KEY
ON KEY LABEL CTRL-F9 SUSPEND

STORE SPACE(0) TO lc_File1, ;
                  lc_File2, ;
                  lc_OutputFile, ;
                  lc_Field1, ;
                  lc_Field2, ;
                  lc_Alias1, ;
                  lc_Alias2, ;
                  lc_IgnoreFieldList

STORE 0 TO ln_Count, ;
           ln_Remove, ;
           ln_ErrorCount

lc_Deleted = SET("DELETED")

STORE .t. TO ll_Cont, ;
             ll_FirstRec

IF NOT pl_TimeStampCompare
  *-- Set files to ignore
  lc_IgnoreFieldList = "ADDUSER, ADDDATE, ADDTIME, LCKUSER, LCKDATE, LCKTIME"
ENDIF

IF TYPE("pc_IgnoreFieldList") = "C" AND ;
  NOT EMPTY(pc_IgnoreFieldList)
  *-- Add parameter files to ignore 

  IF NOT EMPTY(lc_IgnoreFieldList)
    lc_IgnoreFieldList = lc_IgnoreFieldList + ", "
  ENDIF

  lc_IgnoreFieldList = lc_IgnoreFieldList + pc_IgnoreFieldList
ENDIF

IF NOT EMPTY(lc_IgnoreFieldList)
  *-- Remove spaces
  lc_IgnoreFieldList = STRTRAN(lc_IgnoreFieldList, SPACE(1))

  lc_IgnoreFieldList = UPPER(lc_IgnoreFieldList) + ","
ENDIF

IF UPPER(VERSION(1)) = "VISUAL FOXPRO 05"
  = MESSAGEBOX("Does not run under Visual FoxPro 5. Needs 6.0 or above.")
  ll_Cont = .f.
ENDIF

IF NOT pl_FileAlias
  *-- Use tables

  IF EMPTY(pn_SqlHandle)

    *-- Begin: For VFP, verify if the files are there
    IF ll_Cont
      lc_File1 = DEFAULTEXT(ALLTRIM(pc_File1), "DBF")
      lc_File2 = DEFAULTEXT(ALLTRIM(pc_File2), "DBF")

      IF NOT FILE(lc_File1)
        = MESSAGEBOX("File not found " + lc_File1)
        ll_Cont = .f.
      ENDIF

    ENDIF

    IF ll_Cont

      IF NOT FILE(lc_File2)
        = MESSAGEBOX("File not found " + lc_File2)
        ll_Cont = .f.
      ENDIF

    ENDIF

    *-- End: For VFP, verify if the files are there

  ELSE
    *-- SQL tables
    lc_File1 = ALLTRIM(pc_File1)
    lc_File2 = ALLTRIM(pc_File2)
  ENDIF

  IF ll_Cont
    lc_Alias1 = "a_tmpFile1"
    lc_Alias2 = "a_tmpFile2"
  ENDIF

ELSE
  *-- Use aliases

  IF ll_Cont

    lc_Alias1 = ALLTRIM(pc_File1)
    lc_Alias2 = ALLTRIM(pc_File2)

    IF NOT USED(lc_Alias1)
      = MESSAGEBOX("Alias not found " + lc_Alias1)
      ll_Cont = .f.
    ENDIF

  ENDIF

  IF ll_Cont

    IF NOT USED(lc_Alias2)
      = MESSAGEBOX("Alias not found " + lc_Alias2)
      ll_Cont = .f.
    ENDIF

  ENDIF

  IF ll_Cont
    lc_File1 = FULLPATH(DBF(lc_Alias1))
    lc_File2 = FULLPATH(DBF(lc_Alias2))
  ENDIF

ENDIF && NOT pl_FileAlias

DO CASE
  CASE NOT ll_Cont
    *-- Do nothing
  CASE EMPTY(pc_OutputFile)
    lc_OutputFile = PUTFILE("Output File", "temp.txt", "*.TXT")

    IF EMPTY(lc_OutputFile)
      ll_Cont = .f.
    ENDIF

  OTHERWISE
    lc_OutputFile = ALLTRIM(pc_OutputFile)
ENDCASE

IF ll_Cont

  IF NOT pl_NoEraseOutputFile
    ERASE (lc_OutputFile)
  ENDIF

  SET DELETED ON

  DO CASE
    CASE pl_FileAlias
      *-- File aliases given, do nothing
    CASE EMPTY(pn_SqlHandle)
      *-- Open VFP tables
      SELECT 0
      USE (lc_File1) ALIAS (lc_Alias1)

      SELECT 0
      USE (lc_File2) ALIAS (lc_Alias2)
    OTHERWISE
      *-- Get SQL table data
      SELECT 0
      ll_Cont= SQLEXEC(pn_SqlHandle, "SELECT * FROM " + lc_File1, lc_Alias1) > 0
      
      IF NOT ll_Cont
        = MESSAGEBOX("Error when opening " + lc_File1 + "." + Cc_OsEol + ;
          MESSAGE())
      ELSE
        SELECT 0
        ll_Cont= SQLEXEC(pn_SqlHandle, "SELECT * FROM " + lc_File2, lc_Alias2) > 0
        
        IF NOT ll_Cont
          = MESSAGEBOX("Error when opening " + lc_File2 + "." + Cc_OsEol + ;
            MESSAGE())
        ENDIF

      ENDIF
        
  ENDCASE

ENDIF

IF ll_Cont
  SELECT (lc_Alias1)
  = AFIELDS(la_Fields)

  IF NOT EMPTY(lc_IgnoreFieldList)
    ln_remove = 0

    *-- Delete adduser etc fields
    FOR ln_Count = 1 TO ALEN(la_Fields, 1)

      IF TYPE("la_Fields[ln_Count, 1]") <> "C"
        EXIT
      ENDIF

      IF UPPER(la_Fields[ln_Count, 1]) + "," $ lc_IgnoreFieldList
        = ADEL(la_Fields, ln_Count)
        ln_Count = ln_Count - 1
        ln_Remove = ln_Remove + 1
      ENDIF

    NEXT

    IF ln_Remove = ALEN(la_Fields, 1)
      = MESSAGEBOX("No fields to compare.")
      ll_Cont = .f.
    ELSE
      DECLARE la_Fields[ALEN(la_Fields, 1) - ln_Remove, ALEN(la_Fields, 2)]
    ENDIF

  ENDIF

ENDIF && ll_Cont

IF ll_Cont
  = STRTOFILE("DBF compare utility :" + ;
    Cc_OsEol, ;
    lc_OutputFile, .t.)
  = STRTOFILE("File 1 : " + lc_File1 + ;
    Cc_OsEol, ;
    lc_OutputFile, .t.)
  = STRTOFILE("File 2 : " + lc_File2 + ;
    Cc_OsEol, ;
    lc_OutputFile, .t.)

  IF TYPE("pc_Description") = "C" AND ;
    NOT EMPTY(pc_Description)
    = STRTOFILE(pc_Description + ;
      Cc_OsEol, ;
      lc_OutputFile, .t.)
    lc_Description = pc_Description
  ELSE
    lc_Description = ""
  ENDIF

  = STRTOFILE(REPLICATE("-", 80) + ;
    Cc_OsEol, ;
    lc_OutputFile, .t.)

  *-- Loop through the primary table
  SCAN REST

    WAIT WINDOW PADR(lc_Description, 50) + Cc_OsEol + ;
      "Current record : " + ALLTRIM(STR(RECNO(lc_Alias1), 10)) ;
      NOWAIT NOCLEAR

    IF EOF(lc_Alias2)
      = STRTOFILE("There are more records in File1 than file 2" + ;
        Cc_OsEol, ;
        lc_OutputFile, .t.)
      ln_ErrorCount = ln_ErrorCount + 1
      EXIT
    ENDIF

    FOR ln_Count = 1 TO ALEN(la_Fields, 1)
      lc_Field = la_Fields[ln_Count, 1]

      lc_Field1 = lc_Alias1 + "->" + la_Fields[ln_Count, 1]
      lc_Field2 = lc_Alias2 + "->" + la_Fields[ln_Count, 1]

      DO CASE
        CASE TYPE(lc_Field1) = "G"
          *-- General field. Ignore it
        CASE TYPE(lc_Field2) = "U"
          *-- Field not found in 2nd file. Ignore it
        CASE TYPE(lc_Field1) <> TYPE(lc_Field2)
          *-- Fields are of different type. Ignore it
        CASE EVALUATE(lc_Field1) <> EVALUATE(lc_Field2) OR ;
          EVALUATE(lc_Field2) <> EVALUATE(lc_Field1)
          *-- Not an exact match. Report it.
          = STRTOFILE("Record " + STR(RECNO(lc_Alias1), 10) + " : ", ;
            lc_OutputFile, .t.)

          IF NOT EMPTY(pc_RecordExpr)
            = STRTOFILE(ALLTRIM(TRANSFORM(EVALUATE(pc_RecordExpr))), ;
              lc_OutputFile, .t.)
          ENDIF

          = STRTOFILE(" : " + "Field " + la_Fields[ln_Count, 1] + " does not match" + ;
            Cc_OsEol, ;
            lc_OutputFile, .t.)

          = STRTOFILE(SPACE(4) + [Values : "] + ;
            TRANSFORM(EVALUATE(lc_Field1)) + ["] + ;
            Cc_OsEol, ;
            lc_OutputFile, .t.)
          = STRTOFILE(SPACE(4) + SPACE(LEN("Values : ")) + ["] + ;
            TRANSFORM(EVALUATE(lc_Field2)) + ["] + ;
            Cc_OsEol, ;
            lc_OutputFile, .t.)
          ln_ErrorCount = ln_ErrorCount + 1
      ENDCASE

    NEXT

    SKIP IN (lc_Alias2)
  ENDSCAN

  IF NOT EOF(lc_Alias2)
    = STRTOFILE("There are more records in File2 than file 1" + ;
      Cc_OsEol, ;
      lc_OutputFile, .t.)
    ln_ErrorCount = ln_ErrorCount + 1
  ENDIF

  IF ln_ErrorCount > 0
    = STRTOFILE("** Total Differences : " + STR(ln_ErrorCount, 10, 0) + ;
      Cc_OsEol, ;
      lc_OutputFile, .t.)
  ELSE
    = STRTOFILE("^^ No Differences found ^^" + ;
      Cc_OsEol, ;
      lc_OutputFile, .t.)
    ln_ErrorCount = ln_ErrorCount + 1
  ENDIF

  = STRTOFILE(REPLICATE("=", 80) + ;
    Cc_OsEol, ;
    lc_OutputFile, .t.)
ENDIF && ll_Cont

IF NOT ll_Cont
  lc_Outputfile = Cc_Null
ENDIF

WAIT CLEAR

IF ll_Cont AND ;
  NOT pl_NoDisplayErrors AND ;
  FILE(lc_OutputFile)

  MODIFY FILE (lc_OutputFile) NOEDIT
ENDIF

SET DELETED &lc_Deleted.

IF NOT pl_FileAlias

  IF USED(lc_Alias1)
    USE IN (lc_Alias1)
  ENDIF

  IF USED(lc_Alias2)
    USE IN (lc_Alias2)
  ENDIF

ENDIF

POP KEY

RETURN lc_OutputFile
*------------------------------------------------------------------------------
*- File Name..: $Workfile: DBFCMPRE.PRG $
*- Total Lines: 444
